from threading import Thread
import sys
import cv2
import time
from queue import Queue
import numpy as np


class VideoStream:
    
    def __init__(self, screen_size, transform=None, queue_size=128):
        # initialize the file video stream along with the boolean
        # used to indicate if the thread should be stopped or not
        self.stopped = False
        self.transform = transform
        # initialize the queue used to store frames 
        self.Q = Queue(maxsize=queue_size)
        # intialize thread
        self.thread = Thread(target=self.__update, args=())
        self.thread.daemon = True
        self.border_size = 20
        self.screen_size = screen_size
        self.screen_data = np.zeros((screen_size[1]+1, screen_size[0]+1, 3), dtype=np.uint8)

    def update_screen(self, new_screen_data):
        # change the data displayed on the screen
        self.screen_data = new_screen_data

    def get_screen(self):
        # get the current screen data
        return self.screen_data

    def start(self):
        # start a thread to read frames from the file video stream
        self.thread.start()
        return self

    def __update(self):
        # keep looping infinitely
        while True:
            # if the thread indicator variable is set, stop the
            # thread
            if self.stopped:
                break

            # otherwise, ensure the queue has room in it
            if not self.Q.full():
                frame = self.screen_data
                self.Q.put(frame)
            else:
                time.sleep(0.1)  # Rest for 10ms, we have a full queue

    # Insufficient to have consumer use while(more()) which does
    # not take into account if the producer has reached end of
    # file stream.
    def __running(self):
        return self.__more() or not self.stopped

    def __more(self):
        # return True if there are still frames in the queue. If stream is not stopped, try to wait a moment
        tries = 0
        while self.Q.qsize() == 0 and not self.stopped and tries < 5:
            time.sleep(0.1)
            tries += 1
        return self.Q.qsize() > 0

    def stop(self):
        # indicate that the thread should be stopped    
        self.stopped = True
        # wait until stream resources are released (producer thread might be still grabbing frame)
        self.thread.join()


# RM Nimbus colour tables.  

# The Nimbus provided 16 colours in low-resolution mode.  In high
# resolution mode only 4 colours were available, although you can
# assign any of the 16 colours to the 4 slots in the palette.

# The full colour table here is also used in low-res mode.  The key
# is the Nimbus colour number and the value is the BGR composition.
colour_table = {
    #     B    G    R
    0:  [  0,   0,   0],    # black
    1:  [170,   0,   0],    # dark blue
    2:  [  0,   0, 170],    # dark red
    3:  [170,   0, 170],    # purple
    4:  [  0, 170,   0],    # dark green
    5:  [170, 170,   0],    # dark cyan
    6:  [  0,  84, 170],    # brown
    7:  [170, 170, 170],    # light grey
    8:  [ 84,  84,  84],    # dark grey
    9:  [255,  84,  84],    # light blue
    10: [ 84,  84, 255],    # light red
    11: [255,  84, 255],    # light purple
    12: [ 84, 255,  84],    # light green
    13: [255, 255,  84],    # light cyan
    14: [ 84, 255, 255],    # yellow
    15: [255, 255, 255]     # white
}

# Low-res default colours
low_res_default_colours = {
    0:  0,    # black
    1:  1,    # dark blue
    2:  2,    # dark red
    3:  3,    # purple
    4:  4,    # dark green
    5:  5,    # dark cyan
    6:  6,    # brown
    7:  7,    # light grey
    8:  8,    # dark grey
    9:  9,    # light blue
    10: 10,    # light red
    11: 11,    # light purple
    12: 12,    # light green
    13: 13,    # light cyan
    14: 14,    # yellow
    15: 15    # white
}

# High-res default colours
high_res_default_colours = {
    0: 1,                   # dark blue
    1: 4,                   # dark green
    2: 10,                  # light red
    3: 15                   # white
}



from .tools import message, fatal, is_valid_colour, colour_to_bgr, fix_coord, colrows_to_xy
from .tools import plonk_image, plonk_transparent_image, colour_char
import numpy as np
import cv2
import time

class Command:
    """Nimbus commands

    This class contains the commands to control the Nimbus display.  The commands
    mimick the syntax found in RM Basic but re-written Pythonically.  In creating an 
    object of this class it must be bound to a pre-existing Nimbus object.  It is
    highly recommended to read the original Nimbus manuals (particularly RM Basic)
    to get a deeper understanding of how these commands originally worked.

    Args:
        nimbus (Nimbus): The Nimbus object to bind to

    """

    def __init__(self, nimbus):
        """Create a new Command object

        In creating an object of this class it must be bound to
        a pre-existing Nimbus object.

        Args:
            nimbus (Nimbus): The Nimbus object to bind to
        
        """

        # Validate params
        # To avoid circular import we can't put Nimbus in isintance but we can
        # assert that the object contains a string called title.
        assert isinstance(nimbus.title, str), "Command object needs to be bound to a Nimbus object"
        self.nimbus = nimbus


    def cls(self):
        """Clear the screen of all text and graphics and reset cursor position

        """

        # Wipe screen data in the Nimbus and fill screen with paper colour. First
        # define a new PIL image to make sure it matches the current screen mode, 
        # e.g. if set_mode was just called
        screen_data = np.zeros((self.nimbus.screen_size[1]+1, self.nimbus.screen_size[0]+1, 3), dtype=np.uint8)
        cv2.rectangle(screen_data, (0,0), (self.nimbus.screen_size[0], self.nimbus.screen_size[1]), colour_to_bgr(self.nimbus, self.nimbus.paper_colour),-1)
        self.nimbus.update_screen(screen_data)


    def set_mode(self, columns):
        """Select either high-resolution or low-resolution screen mode

        In RM Basic the screen resolution was set by the number of columns: 
        40 for low-resolution and 80 for high-resolution.  Any other values
        has no effect.  Nimbusinator is more strict and will yield an error
        if any other values are entered.  Check the original RM Basic manual
        for a description of how screen resolutions worked on the Nimbus.
        High-resolution mode is particularly odd.

        Args:
            columns (int): The number of colums (40 or 80)

        """
        
        # validate params
        assert isinstance(columns, int), "The value of columns must be integer, not {}".format(columns)
        assert (columns == 80 or columns == 40), "The value of columns can be 80 or 40, not {}".format(columns)

        # change screen size and set default colours accordingly
        if columns == 80:
            self.nimbus.screen_size = (640, 250)
            self.nimbus.high_res_colours = self.nimbus.high_res_default_colours.copy()
            self.set_paper(0)
            self.set_border(0)
            self.set_brush(3)
            self.set_pen(3)
            self.set_curpos((1, 1))
            self.cls()
            return
        if columns == 40:
            self.nimbus.screen_size = (320, 250)
            self.nimbus.low_res_colours = self.nimbus.low_res_default_colours.copy()
            self.set_paper(0)
            self.set_border(0)
            self.set_brush(15)
            self.set_pen(15)
            self.set_curpos((1, 1))
            self.cls()
            return


    def set_cursor(self, flag):
        """Show or hide cursor

        Args:
            flag (boolean): True to show cursor, False to hide

        """
        
        # validate params
        assert isinstance(flag, bool), "The value of flag must be boolean, not {}".format(flag)

        self.nimbus.show_cursor = flag


    def plonk_logo(self, coord):
        """Plonk the RM Nimbus logo on screen

        Args:
            coord (tuple): The (x, y) position to plonk the logo

        """

        # validate params
        assert isinstance(coord, tuple), "The value of coord must be a tuple, not {}".format(coord)
        assert len(coord) == 2, "The coord tuple must have 2 values, not {}".format(len(coord))
        for i in range(0, 2):
            assert isinstance(coord[i], int), "This value in coord must be an integer, not{}".format(coord[i])

        screen_data = self.nimbus.get_screen()
        screen_data = plonk_image(self.nimbus, screen_data, self.nimbus.logo, coord)
        self.nimbus.update_screen(screen_data)


    def set_paper(self, colour):
        """Set the paper colour
        
        Args:
            colour (int): Colour value (High-resolution: 0-3, low-resolution: 0-15)

        """

        # validate params
        assert isinstance(colour, int), "The value of colour must be an integer, not {}".format(type(colour))
        assert (colour == is_valid_colour(self.nimbus, colour)), "Colour {} is out-of-range for this screen mode".format(colour)

        self.nimbus.paper_colour = colour
        
    
    def set_colour(self, colour1, colour2):
        """Set a colour to a new colour

        Args:
            colour1 (int): The colour code to be changed
            colour2 (int): The new colour to be assigned to colour1
        
        """

        # validate params
        assert isinstance(colour1, int), "The value of colour1 must be an integer, not {}".format(type(colour1))
        assert isinstance(colour2, int), "The value of colour2 must be an integer, not {}".format(type(colour2))
        assert (colour2 >= 0 and colour2 <= 15), "The value of colour 2 must be >= 0 and <= 15, not {}".format(colour2)
        assert (colour1 == is_valid_colour(self.nimbus, colour1)), "Colour1 {} is out-of-range for this screen mode".format(colour1)

        if self.nimbus.screen_size == (320, 250):
            self.nimbus.low_res_colours[colour1] = colour2
        if self.nimbus.screen_size == (640, 250):
            self.nimbus.high_res_colours[colour1] = colour2


    def set_border(self, colour):
        """Set the border colour
        
        Args:
            colour (int): Colour value (High-resolution: 0-3, low-resolution: 0-15)
            
        """
 
        # validate params
        assert isinstance(colour, int), "The value of colour must be an integer, not {}".format(type(colour))
        assert (colour == is_valid_colour(self.nimbus, colour)), "Colour {} is out-of-range for this screen mode".format(colour)

        self.nimbus.border_colour = colour
    

    def set_brush(self, colour):
        """Set the brush colour
        
        Args:
            colour (int): Colour value (High-resolution: 0-3, low-resolution: 0-15)
            
        """

        # validate params
        assert isinstance(colour, int), "The value of colour must be an integer, not {}".format(type(colour))
        assert (colour == is_valid_colour(self.nimbus, colour)), "Colour {} is out-of-range for this screen mode".format(colour)

        self.nimbus.brush_colour = colour


    def set_pen(self, colour):
        """Set the pen colour

        Args:
            colour (int): Colour value (High-resolution: 0-3, low-resolution: 0-15)
            
        """

        # validate params
        assert isinstance(colour, int), "The value of colour must be an integer, not {}".format(type(colour))
        assert (colour == is_valid_colour(self.nimbus, colour)), "Colour {} is out-of-range for this screen mode".format(colour)

        self.nimbus.pen_colour = colour


    def set_charset(self, charset):
        """Set the charset for text

        Args:
            charset (int): 0 is the standard font, 1 is the other font!

        """

        # validate params
        assert isinstance(charset, int), "The value of charset must be an integer, not {}".format(type(charset))
        assert (charset == 0 or charset == 1), "The value of charset can be either 0 or 1, not {}".format(charset)

        self.nimbus.charset = charset


    def ask_charset(self):
        """Return the current charset for text

        Returns:
            charset (int): 0 is the standard font, 1 is the other font

        """

        return self.nimbus.charset


    def set_curpos(self, cursor_position):
        """Set the cursor position

        Args:
            cursor_position (tuple): The new cursor position (column, row)

        """

        # Validate params
        assert isinstance(cursor_position, tuple), "The value of cursor_position must be a tuple, not {}".format(type(cursor_position))
        assert len(cursor_position) == 2, "The cursor_position tuple must have 2 values, not {}".format(len(cursor_position))
        for i in range(0, 2):
            assert isinstance(cursor_position[i], int), "This value in cursor_position must be an integer, not{}".format(type(cursor_position[i]))
        # Validate that cursor will still be on screen
        assert (colrows_to_xy(self.nimbus.screen_size, (cursor_position[0], 1))[0] < self.nimbus.screen_size[0]), "Column {} is not on the screen".format(cursor_position[0])
        assert (colrows_to_xy(self.nimbus.screen_size, (1, cursor_position[1]))[1] >= 0), "Row {} is not on the screen".format(cursor_position[1])
        assert (cursor_position[0] >= 0), "Negative column value in {} is not permitted".format(cursor_position)
        assert (cursor_position[1] >= 0), "Negative row value in {} is not permitted".format(cursor_position)       

        # Now update the cursor position
        self.nimbus.update_cursor_position(cursor_position)


    def ask_curpos(self):
        """Gets the current cursor position

        Returns:
            cursor_position (tuple): The current cursor position (column, row)

        """

        # Return cursor position
        return self.nimbus.get_cursor_position()


    def plot(self, text, coord, size=1, brush=None, direction=0, font=None):
        """Plot text on the screen

        Args:
            text (str): The text to be plotted
            coord (tuple): The (x, y) position of the text
            size (int), optional: Font size. To elongate pass a tuple (x_size, y_size)
            brush (int), optional: Brush colour
            direction (int), optional: 0=normal, 1=-90deg, 2=180deg, 3=-270deg
            font (int), optional: 0 is the standard font, 1 is the other font

        """

        # Validate params
        assert isinstance(text, str), "The value of text must be a string, not {}".format(type(str))
        assert isinstance(coord, tuple), "The value of coord must be a tuple, not {}".format(type(coord))
        assert len(coord) == 2, "The coord tuple must have 2 values, not {}".format(len(coord))
        for i in range(0, 2):
            assert isinstance(coord[i], int), "The values in coord {} must be integer, not {}".format(coord, type(coords[i]))      
        assert isinstance(size, (int, tuple)), "The value of size must be an integer or tuple, not {}".format(type(size))
        if isinstance(size, tuple):
            for i in range(0, 2):
                assert isinstance(size[i], int), "The values in size {} must be integer, not {}".format(size, type(size[i]))  
        assert isinstance(brush, (type(None), int)), "The value of brush must be None or an integer, not {}".format(type(brush))
        assert (brush == is_valid_colour(self.nimbus, brush)), "Brush colour {} is out-of-range for this screen mode".format(brush)
        assert isinstance(direction, int), "The value of direction must be an integer, not {}".format(type(direction))
        assert isinstance(font, (type(None), int)), "The value of font must be an integer, not {}".format(type(font))
        assert (font == 0 or font == 1 or font is None), "The value of font can be 0 or 1, not {}".format(font)
        

        # Handle is_black workaround
        if colour_to_bgr(self.nimbus, brush) == [0, 0, 0]:
            is_black = True
        else:
            is_black = False

        # Handle brush colour
        if brush is None:
            brush = self.nimbus.brush_colour

        # Handle font
        if font is None:
            font = self.nimbus.plot_font

        # Create a temporary image of the plotted text
        plot_img_width = len(text) * 9
        plot_img = np.zeros((10, plot_img_width, 3), dtype=np.uint8)
        x = 0
        for char in text:
            if is_black:
                char_img = cv2.bitwise_not(self.nimbus.font_images[font][ord(char)])
            else:
                char_img = colour_char(self.nimbus, brush, cv2.bitwise_not(self.nimbus.font_images[font][ord(char)]))
            plot_img = plonk_image(self.nimbus, plot_img, char_img, (x, 0), custom_size=(plot_img_width, 10))
            x += 8
        
        # resize
        if isinstance(size, tuple):
            # tuple: extract x_size, y_size
            x_size, y_size = size
        else:
            x_size = size
            y_size = size
        resized = cv2.resize(plot_img, (plot_img.shape[1]*x_size, plot_img.shape[0]*y_size), interpolation=0)
        
        # rotate
        for i in range(0, direction):
            resized = cv2.rotate(resized, cv2.ROTATE_90_COUNTERCLOCKWISE)
        
        # rebuild screen and done
        screen_data = self.nimbus.get_screen()
        screen_data = plonk_transparent_image(self.nimbus, screen_data, resized, coord, is_black=is_black)
        self.nimbus.update_screen(screen_data)


    def flush(self):
        """Clears the keyboard buffer

        """

        self.nimbus.keyboard_buffer = []
    

    def gets(self):
        """Get the oldest char in the keyboard buffer

        Equivalent to GET$ in RM Basic

        Returns:
            (str): The oldest char in the buffer
        
        """

        # If the buffer isn't empty pop the last char
        # and return it, otherwise return empty str
        if len(self.nimbus.keyboard_buffer) > 0:
            return self.nimbus.keyboard_buffer.pop(0)
        else:
            return ''


    def input(self, prompt):
        """Collects keyboard input until user presses ENTER

        Args:
            prompt (str): Message to be printed

        Returns:
            (str): The user's response
        
        """
        
        # Validata params
        assert isinstance(prompt, str), "The value of prompt must be a string, not {}".format(type(prompt))

        # Get max columns for this screen mode
        if self.nimbus.screen_size == (320, 250):
            max_columns = 40
        if self.nimbus.screen_size == (640, 250):
            max_columns = 80
        # Flush buffer, reset enter + delete flag
        self.flush()
        self.nimbus.enter_was_pressed = False
        self.nimbus.backspace_was_pressed = False
        # Print the prompt and get start cursor position
        self.put(prompt)
        # Collect response in this string:
        response = ''
        # Collect and echo chars from buffer until enter was pressed
        while not self.nimbus.enter_was_pressed:
            new_char = self.gets()
            response += new_char
            self.put(new_char)
            # Handle delete
            if self.nimbus.backspace_was_pressed and len(response) == 0:
                self.nimbus.backspace_was_pressed = False
            if self.nimbus.backspace_was_pressed and len(response) > 0:
                now_col, now_row = self.ask_curpos()
                response = response[:-1]
                # Move cursor left
                # If we're about to move off the left-hand side of the screen
                # it must be because we're on a line below where the input
                # started.  So, we need to move up one row and locate cursor
                # at right-hand side, wipe whatever char is there, then re-
                # position
                next_col = now_col - 1
                if next_col == 0:
                    # Go up one row
                    next_col = max_columns
                    next_row = now_row - 1
                else:
                    # Go back one column
                    next_row = now_row
                # Wipe char in this location, reposition, and reset flag
                self.set_curpos((next_col, next_row))
                self.put(' ')
                self.set_curpos((next_col, next_row))
                self.nimbus.backspace_was_pressed = False
        # Enter was pressed, flush buffer and reset enter flag
        self.nimbus.enter_was_pressed = False
        self.flush()
        # Force carriage return by smashing the cursor off the screen
        col, row = self.ask_curpos()
        self.nimbus.cursor_position = (255, row)
        self.put('X')
        # Return string
        return response


    def put(self, ascii_data):
        """Put a single character or string at the current cursor position

        Args:
            ascii_data (int/str): If an int is passed the corresponding ASCII character
                                    will be plotted.  If a string is passed then the 
                                    string will be printed without a terminating carriage
                                    return.

        """
        
        # Validate params
        assert isinstance(ascii_data, (str, int)), "The value of ascii_data must be an integer or string, not {}".format(type(ascii_data))

        # Handle is_black workaround
        if colour_to_bgr(self.nimbus, self.nimbus.pen_colour) == [0, 0, 0]:
            is_black = True
        else:
            is_black = False

        # Handle integer
        if isinstance(ascii_data, int):
            # validate in extended ASCII range
            assert (ascii_data >= 0 and ascii_data <= 255), "The value {} of ascii_data is outside the range of Extended ASCII (0-255)".format(ascii_data)
            # convert to char
            ascii_list = [chr(ascii_data)]
        # Handle string
        if isinstance(ascii_data, str):
            ascii_list = ascii_data
        
        # Put char or chars
        for ascii in ascii_list:
            # Get char img
            char_img = self.nimbus.font_images[self.nimbus.charset][ord(ascii)]
            # Get screen position in pixels from cursor position
            curpos_xy = colrows_to_xy(self.nimbus.screen_size, self.nimbus.get_cursor_position())
            # Plot char and apply paper colour underneath char
            screen_data = self.nimbus.get_screen()
            # Paper colour
            cv2.rectangle(
                screen_data, 
                fix_coord(self.nimbus.screen_size, (curpos_xy[0], curpos_xy[1])), 
                fix_coord(self.nimbus.screen_size, (curpos_xy[0] + 8, curpos_xy[1] + 10)), 
                colour_to_bgr(self.nimbus, self.nimbus.paper_colour), 
                -1
            )
            # Overlay char, colourise and preserve paper colour
            if is_black:
                char_img = cv2.bitwise_not(char_img)
            else:
                char_img = colour_char(self.nimbus, self.nimbus.pen_colour, cv2.bitwise_not(char_img))
            screen_data = plonk_transparent_image(self.nimbus, screen_data, char_img, (curpos_xy[0], curpos_xy[1]), is_black=is_black)
            # calculate new curpos, if over the right-hand side do carriage return
            self.nimbus.update_screen(screen_data)
            new_column = self.nimbus.get_cursor_position()[0] + 1
            if colrows_to_xy(self.nimbus.screen_size, (new_column, 1))[0] >= self.nimbus.screen_size[0]:
                # do carriage return
                new_column = 1  # return to left-hand side
                new_row = self.nimbus.get_cursor_position()[1] + 1  # move down
                # if we're below, then screen, move screen data up 10 pixels and set
                # cursor to bottom of screen
                if colrows_to_xy(self.nimbus.screen_size, (1, new_row))[1] < 0:
                    new_row = self.nimbus.get_cursor_position()[1]
                    # Shove screen up.  First crop the top line:
                    old_screen_data = self.nimbus.get_screen()[10:, :]
                    # Make a blank screen and apply paper colour (same as Nimbus did it)
                    screen_data = np.zeros((self.nimbus.screen_size[1]+1, self.nimbus.screen_size[0]+1, 3), dtype=np.uint8)
                    cv2.rectangle(screen_data, (0,0), (self.nimbus.screen_size[0], self.nimbus.screen_size[1]), colour_to_bgr(self.nimbus, self.nimbus.paper_colour),-1)
                    # And overlay the old_screen_data
                    screen_data[:-10, :] = old_screen_data
                    # Update screen
                    self.nimbus.update_screen(screen_data)
            else:
                # don't move cursor down
                new_row = self.nimbus.get_cursor_position()[1]
            # move cursor
            self.set_curpos((new_column, new_row))
         
            
    def print(self, text):
        """Print a string with carriage return at end

        Args:
            text (str): The text to be printed

        """

        # Validate params
        assert isinstance(text, str), "The value of text must be a string, not {}".format(type(text))

        # Put the string and then maybe CR
        self.put(text)
        # Carriage return?
        col, row = self.ask_curpos()
        if col > 1:
            # Yep - smash the cursor off the screen and use put to force CR
            self.nimbus.cursor_position = (255, row)
            self.put('X')


    def line(self, coord_list, brush=None):
        """Draw one or more connected straight lines

        Args:
            coord_list (list): A list of (x, y) tuples
            brush (int), optional: Colour value (High-resolution: 0-3, low-resolution: 0-15)

        """

        # validate params
        assert isinstance(coord_list, list), "coord_list should contain a list, not {}".format(type(coord_list))
        assert isinstance(brush, (type(None), int)), "The value of brush must be None or an integer, not {}".format(type(brush))
        assert (brush == is_valid_colour(self.nimbus, brush)), "Brush colour {} is out-of-range for this screen mode".format(brush)
        for coord in coord_list:
            assert len(coord) == 2, "The coord tuple must have 2 values, not {}".format(len(coord))
            for i in range(0, 2):
                assert isinstance(coord[i], int), "The values in coord {} must be integer, not {}".format(coord, type(coord[i])) 

        # if default brush value then get current brush colour
        if brush is None:
            brush = self.nimbus.brush_colour
        # validate brush
        brush = is_valid_colour(self.nimbus, brush)
        # draw lines on screen
        screen_data = self.nimbus.get_screen()
        for i in range(0, len(coord_list) - 1):
            cv2.line(screen_data, fix_coord(self.nimbus.screen_size, coord_list[i]), fix_coord(self.nimbus.screen_size, coord_list[i+1]), colour_to_bgr(self.nimbus, brush), 1)
        self.nimbus.update_screen(screen_data)
    

    def area(self, coord_list, brush=None):
        """Draw a filled polygon

        Args:
            coord_list (list): A list of (x, y) tuples
            brush (int), optional: Colour value (High-resolution: 0-3, low-resolution: 0-15)

        """

        # validate params
        assert isinstance(coord_list, list), "coord_list should contain a list, not {}".format(type(coord_list))
        assert isinstance(brush, (type(None), int)), "The value of brush must be None or an integer, not {}".format(type(brush))
        assert (brush == is_valid_colour(self.nimbus, brush)), "Brush colour {} is out-of-range for this screen mode".format(brush)
        for coord in coord_list:
            assert len(coord) == 2, "The coord tuple must have 2 values, not {}".format(len(coord))
            for i in range(0, 2):
                assert isinstance(coord[i], int), "The values in coord {} must be integer, not {}".format(coord, type(coord[i]))

        # if default brush value then get current brush colour
        if brush is None:
            brush = self.nimbus.brush_colour
        # validate brush
        brush = is_valid_colour(self.nimbus, brush)
        # convert coord_list into array
        poly_list = []
        for coord in coord_list:
            coord = fix_coord(self.nimbus.screen_size, coord)
            poly_list.append([coord[0], coord[1]])
        # draw filled polygon on screen
        screen_data = self.nimbus.get_screen()
        cv2.fillPoly(screen_data, np.array([poly_list]), colour_to_bgr(self.nimbus, brush))
        self.nimbus.update_screen(screen_data)


from .tools import logo, message, fatal, colour_to_bgr, font_image_selecta, colrows_to_xy, plonk_image
from .command import Command
from .videostream import VideoStream
from .colour_table import colour_table, low_res_default_colours, high_res_default_colours
from .welcome import welcome
import cv2
import numpy as np
import time
import threading
import random
import simpleaudio as sa
from pynput import keyboard
import os


# get full path of this script
real_path = os.path.dirname(os.path.realpath(__file__))

# define default cursor image
default_cursor_image = np.ones((2, 10, 3), dtype=np.uint8) * 255


class Nimbus:
    """Nimbus video display class.

    This class represents the Nimbus video display that will host the user
    interface for your application.  When created, the new Nimbus object 
    will not be visible until the boot() method has been called.

    Args:
        full_screen (bool), optional: Full screen mode
        title (str), optional: The title of the display window

    """

    def __onMouse(self, event, x, y, flags, param):
        """Handle mouse event

        """


        return


    def __init__(self, full_screen=False, title='Nimbusinator', border_size=40):
        """Create a new Nimbus object

        When created, the new Nimbus object  will
        not be visible until the boot() method has been called.

        Args:
            full_screen (bool), optional: Full screen mode
            title (str), optional: The title of the display window

        """

        # Show logo in console
        print(logo)

        # Validate params
        assert isinstance(full_screen, bool), "value of full_screen must be boolean, not {}".format(full_screen)
        assert isinstance(title, str), "value of title must be string, not {}".format(title)
        assert isinstance(border_size, int), "value of border_size must be integer, not {}".format(border_size)
        assert (border_size >= 0 and border_size <= 100), "border_size must >= 0 and <= 100, not {}".format(border_size)

        # Define class params
        self.full_screen = full_screen                      # Full screen flag
        self.running = False                                # Flag to run or stop the Nimbus
        self.title = title                                  # Display window title
        self.font_images = self.__load_fonts()              # Font images
        logo_path = os.path.join(real_path, 'data', 'rm-nimbus-logo.png')
        self.logo = cv2.imread(logo_path)                   # Nimbus logo image
        self.screen_size = (640, 250)                       # Screen size (initializes in high-res mode)
        self.border_size = border_size                      # Border size (min 0, max 100)
        self.border_colour = 0                              # High-res initial border colour is blue
        self.paper_colour = 0                               # High-res initial paper colour is blue
        self.brush_colour = 3                               # High-res initial brush colour is white
        self.pen_colour = 3                                 # High-res initial pen colour is white
        self.cursor_position = (1, 1)                       # Initial cursor position is top-left
        self.plot_font = 0                                  # Initial plot font (charset for plot)
        self.charset = 0                                    # Initial charset (font for text)
        self.colour_table = colour_table                    # Dict to to convert Nimbus colour numbers to BGR
        self.low_res_default_colours = low_res_default_colours.copy()       # Low-res default colour table
        self.high_res_default_colours = high_res_default_colours.copy()     # High-res default colour table
        self.low_res_colours = low_res_default_colours.copy()               # Editable low-res colour table
        self.high_res_colours = high_res_default_colours.copy()             # Editable high-res colour table
        self.cursor_image = default_cursor_image.copy()     # Text cursor image
        self.cursor_flash = False                           # Cursor flash flag
        self.show_cursor = False                            # Show cursor flag
        self.floppy_is_running = False                      # Floppy drive running flag
        self.keyboard_buffer = []                           # Keyboard buffer
        self.ctrl_pressed = False                           # CTRL is pressed flag
        self.enter_was_pressed = False                      # ENTER was pressed flag
        self.backspace_was_pressed = False                  # BACKSPACE was pressed flag
        self.__vs = VideoStream(self.screen_size, queue_size=16).start()  # VideoStream object to display the Nimbus


    def __load_fonts(self):
        """Load fonts from PNG files

        Returns:
            (dict): A dict of PIL images for font 0 and font 1

        """

        fonts = {}
        for font in range(0, 2):
            font_img_path = os.path.join(real_path, 'data', 'font{}.png'.format(font))
            font_img = cv2.imread(font_img_path)
            fonts[font] = []
            for ascii_code in range(0, 256):
                fonts[font].append(font_image_selecta(font_img, ascii_code, font))
        return fonts


    def get_cursor_position(self):
        """Get current cursor position

        Returns:
            (tuple): The current cursor position (col, row)

        """

        return self.cursor_position


    def update_cursor_position(self, new_cursor_position):
        """Update cursor position

        Args:
            new_cursor_position (tuple): The new cursor position (col, row)

        """

        self.cursor_position = new_cursor_position


    def update_screen(self, new_screen_data):
        """Update screen data

        Change the screen data to be displayed in VideoStream

        Args:
            new_screen_data (PIL image): The new screen data to be displayed

        """

        self.__vs.update_screen(new_screen_data)


    def get_screen(self):
        """Get screen data

        Get the screen data to be displayed from VideoStream

        Returns:
            (PIL image): The screen data to be displayed
        
        """

        return self.__vs.get_screen()


    def __cycle_cursor_flash(self):
        """Cycle the cursor flash

        This needs to run in it's own thread

        """
        while self.running:
            # When the Nimbus is boot flip cursor flash every 0.5 secs
            time.sleep(0.5)
            cursor_flash = self.cursor_flash
            self.cursor_flash = not cursor_flash
        return


    def __render_display(self, screen_data):
        """Generate final display data including border

        In low-resolution mode the screen size is doubled and maintanes the same
        aspect ratio. It is then centrally overlayed on a larger image that acts
        as a border.  In high-resolution mode the original Nimbus would stretch
        the screen vertically to match the aspect ratio of low-resolution mode,
        making everything appear elongated.  It is then overlayed on the border as
        before.

        Args:
            screen_data (PIL image): The screen data to be displayed

        Returns:
            display_data (PIL image): The finished display data
        """

        # Calculate the actual display dimensions with border:
        horizontal_display_length = 640+(self.border_size*2)
        vertical_display_length = 500+(self.border_size*2)
        # Make the display image as an empty array then add the border colour
        display_data = np.zeros((vertical_display_length, horizontal_display_length, 3), dtype=np.uint8)
        cv2.rectangle(display_data, (0,0), (horizontal_display_length, vertical_display_length), colour_to_bgr(self, self.border_colour), -1)
        # Add cursor if showing cursor
        final_screen_data = screen_data.copy()
        if self.show_cursor:
            coord = colrows_to_xy(self.screen_size, self.cursor_position)
            # Only overlay cursor if on screen and cursor flash is true
            if coord[0] < self.screen_size[0] and self.cursor_flash:
                final_screen_data = plonk_image(self, screen_data, self.cursor_image, coord)
        # resize the combined screen_data and add it to display
        resized = cv2.resize(final_screen_data, (640, 500), interpolation=cv2.INTER_LINEAR_EXACT)
        display_data[self.border_size:self.border_size+resized.shape[0], self.border_size:self.border_size+resized.shape[1]] = resized
        return display_data


    def __screen_runner(self):
        """Display the Nimbus in a window

        This function contains a loop which can be broken by setting
        the running flag to False.  It must be run within its own thread.

        """

        # Set full screen mode in OpenCV
        if self.full_screen:
            cv2.namedWindow(self.title, cv2.WND_PROP_FULLSCREEN)
            cv2.setWindowProperty(self.title, cv2.WND_PROP_FULLSCREEN, cv2.WINDOW_FULLSCREEN)
        cv2.namedWindow(self.title)
        # Nimbus's mouse feature won't be implemented until later
        cv2.setMouseCallback(self.title, self.__onMouse)
        # Display loop
        while self.running:
            frame = self.__render_display(self.__vs.get_screen())
            cv2.imshow(self.title, frame)
            time.sleep(0.05)
            cv2.waitKey(5)
        # Stop cv2 if shutting down
        cv2.destroyAllWindows()
        return


    def __floppy_drive_effects(self):
        """Simulate floppy drive grinding

        This has to be run a thread.  Two drive sounds are supplied: a
        short 'dot' sounds, and a long 'dash' sound, with 4 variations
        of each.  Semi-random patterns of dashes and dots are generated 
        with randomly-selected variations while the floppy_is_running
        flag is True.

        """
        while self.running:
            # Load sound effects
            dash = []
            dot = []
            for i in range(1, 5):
                # Abort if shutting down
                if not self.running:
                    return
                # Otherwise load sound effects
                dash_path = os.path.join(real_path, 'data', 'floppy-dash{}.wav'.format(i))
                dot_path = os.path.join(real_path, 'data', 'floppy-dot{}.wav'.format(i))
                dash.append(sa.WaveObject.from_wave_file(dash_path))
                dot.append(sa.WaveObject.from_wave_file(dot_path))
            # If flag is True play a bunch of grinding floppy drive sounds
            if self.floppy_is_running and self.running:
                # play dash (pick one at random)
                play_obj = dash[random.randint(0, 3)].play()
                play_obj.wait_done()
                # Abort is shutting down
                if not self.running:
                    return
                # Otherwise play random number of dots
                for i in range(0, random.randint(1, 5)):
                    # break out if drive is stopped or shutting down
                    if not self.floppy_is_running or not self.running:
                        return
                    else:
                        # pick one at random
                        play_obj = dot[random.randint(0, 3)].play()
                        play_obj.wait_done()


    def run_floppy(self, flag):
        """Run or stop the floppy drive sound effects

        Augment your user experience with the industrial melodies of a 
        1980s PC floppy drive

        Args:
            flag (boolean): True to run the drive, False to stop

        """

        self.floppy_is_running = flag


    def __on_key_press(self, key):
        """Handle control key presses

        """
        
        # Printable chars go straight into the buffer
        try:
            self.keyboard_buffer.append(key.char)
            # BUT - if CTRL-C situation then shutdown!
            if self.ctrl_pressed and key.char.lower() == 'c':
                message('CTRL-C detected')
                self.shutdown()
        except AttributeError:
            # Handle CTRL released
            if key == keyboard.Key.ctrl_l or keyboard.Key.ctrl_r:
                self.ctrl_pressed = True
            # Handle ENTER hit
            if key == keyboard.Key.enter:
                self.enter_was_pressed = True
            # Handle BACKSPACE hit
            if key == keyboard.Key.backspace:
                self.backspace_was_pressed = True
            # Also add spaces to buffer
            if key == keyboard.Key.space:
                self.keyboard_buffer.append(' ')


    def __on_key_release(self, key):
        """Handle control key releases

        """

        # Handle CTRL released
        if key == keyboard.Key.ctrl_l or keyboard.Key.ctrl_r:
            self.ctrl_pressed = False


    def boot(self, skip_welcome_screen=False):
        """Boot the Nimbus

        Reveal the Nimbus in all its glory, with or without a cheeky
        simulation of the famous Nimbus welcome screen.

        Args:
            skip_welcome_screen (bool), optional: Go straight to the application

        """

        # Validate params
        assert isinstance(skip_welcome_screen, bool), "value of skip_welcome_screen must be boolean, not {}".format(skip_welcome_screen)

        # otherwise go ahead
        message('Press CTRL-C to interrupt')
        message('Booting up')
        # Set running flag
        self.running = True
        # Fire up screen runner in a thread
        self.__t_screen = threading.Thread(target=self.__screen_runner, args=())
        self.__t_screen.start()
        # Fire up cursor in another thread
        self.__t_cursor = threading.Thread(target=self.__cycle_cursor_flash, args=())
        self.__t_cursor.start()
        # Fire up the floppy disk effects in another
        self.__t_floppy = threading.Thread(target=self.__floppy_drive_effects, args=())
        self.__t_floppy.start()
        # Fire up the keyboard listener
        listener = keyboard.Listener(
            on_press=self.__on_key_press,
            on_release=self.__on_key_release)
        listener.start()
        if skip_welcome_screen:
            # don't bother with welcome screen
            Command(self).set_mode(80)
            message('Done')
            return
        else:
            # roll the welcome screen
            welcome(Command(self), self)
            Command(self).set_mode(80)
            message('Done')


    def sleep(self, sleep_time):
        """Pause execution like time.sleep()

        Unlike time.sleep(), this built-in sleep method will be interrupted
        if the user hits CTRL-C.  Sleep time measured in seconds.

        """

        # Validate params
        assert isinstance(sleep_time, (int, float)), "The value of sleep_time must be an integer or float, not {}".format(type(sleep_time))
        assert sleep_time > 0.01, "The value of sleep_time must be > 0.01, not {}".format(sleep_time)

        # Take 10 ms sleeps until sleep_time is reached
        # and check after each sleep if the Nimbus is still running
        elapsed_time = 0
        while elapsed_time <= sleep_time and self.running:
            if not self.running:
                break
            time.sleep(0.01)
            elapsed_time += 0.01
        return


    def shutdown(self):
        """Shutdown the Nimbus

        Stop everything and close the display Window.

        """

        # only shut down if running
        if self.running:
            message('Shutting down')
            message('Press CTRL-C again to exit immediately')
            # Set running flag and join all threads
            self.running = False
            self.__t_cursor.join()
            self.__t_floppy.join()
            self.__t_screen.join()
            raise SystemExit('Exited')
        else:
            raise SystemExit('Exited')


import inspect
import sys
import numpy as np
import cv2


# Let there be ASCII art
logo = """                                        
 _____ _       _           _         _           
|   | |_|_____| |_ _ _ ___|_|___ ___| |_ ___ ___ 
| | | | |     | . | | |_ -| |   | .'|  _| . |  _|
|_|___|_|_|_|_|___|___|___|_|_|_|__,|_| |___|_|  
                                                  
                                               
RM Nimbus GUI for Python
                                            
"""


def message(text):
    """Debug message

    Prints a message in the console including the name of the function
    that sent the message.

    Args:
        text (str): The text of the message

    """

    # Get name of function that sent the message
    caller = inspect.stack()[1][3]
    # Print the debug message
    print('[nimbusinator] {}: {}'.format(caller, text))


def fatal(nimbus):
    """Handle a fatal error

    If a fatal error is encountered shutdown the Nimbus :(

    Args:
        nimbus (Nimbus): The Nimbus object to shutdown
    
    """

    message('Fatal error')
    nimbus.shutdown()
    sys.exit()


def is_valid_colour(nimbus, colour):
    """Validate colour number

    If the colour number is out of range for the current screen
    mode a fatal error is yielded

    Args:
        nimbus (Nimbus): The Nimbus object
        colour (int): The colour number
    
    Returns:
        (int): The validated colour number

    """

    # validate low-res colour
    if nimbus.screen_size == (320, 250):
        if colour >= 0 and colour <= 15:
            # it's fine
            return colour
        else:
            # it's fatal
            return None
    
    # validate high-res colour
    if nimbus.screen_size == (640, 250):
        if colour >= 0 and colour <= 3:
            # it's fine
            return colour
        else:
            # it's fatal
            return None


def colour_to_bgr(nimbus, colour):
    """Convert a validated colour number to BGR composition

    Recommended to use is_valid_colour function first!

    Args:
        nimbus (Nimbus): The Nimbus object
        colour (int): The colour number
    
    Returns:
        (array): The BGR composition in an array, i.e. [B, G, R]
    
    """

    # low-res colour
    if nimbus.screen_size == (320, 250):
        return nimbus.colour_table[nimbus.low_res_colours[colour]]
        #return nimbus.colour_table[colour]
    
    # high-res colour
    if nimbus.screen_size == (640, 250):
        return nimbus.colour_table[nimbus.high_res_colours[colour]]
        #return nimbus.colour_table[nimbus.high_res_colour_table[colour]]


def fix_coord(screen_size, coord):
    """Fix coordinates for use in OpenCV's drawing functions

    PIL images have upside-down co-ordinates and it shafts me
    every goddamn time, so this function "deals with it"

    Args:
        screen_size (tuple): The screen size (width, height)
        coord (tuple): The x, y coordinate to fix (x, y)

    Returns:
        (tuple): The coordinates flipped upside-down

    """
    
    return (coord[0], screen_size[1] - coord[1])


def colrows_to_xy(screen_size, cursor_position):
    """Convert cursor position to x, y pixel position

    Args:
        screen_size (tuple): The screen size (width, height)
        cursor_position (tuple): The cursor position (row, col)

    Returns:
        (tuple): The screen position in pixels (x, y)
    
    """

    x = (8 * (cursor_position[0] - 1))
    y = screen_size[1] - (cursor_position[1] * 10)
    return (x, y)


def ceildiv(a, b):
    return -(-a // b)

def font_image_selecta(font_img, ascii_code, font):
    """Get the image of a character from a PNG

    Enter an ASCII code and a transparent PNG image of the char is returned.
    Codes < 33 and delete char (127) just return a space.

    Args:
        ascii_code (int): The ASCII code (extended) of the character
        font (int): The font or charset

    Returns:
        (PIL image): The transparent image of the character
    
    """

    # On our Nimbus char map PNGs delete (127) is just a blank space, so if we
    # receive any < 33 control chars, set the ascii value to 127 so a space is
    # also returned in those cases.
    if ascii_code < 33:
        ascii_code = 127
    # Calculate the row and column position of the char on the character map PNG
    map_number = ascii_code - 32    # codes < 33 are not on the map (unprintable)
    row = ceildiv(map_number, 30)
    column = map_number - (30 * (row - 1))
    # Calculate corners of box around the char
    x1 = (column - 1) * 10
    x2 = x1 + 10
    y1 = (row - 1) * 10
    y2 = y1 + 10
    # Chop out the char and return as PIL
    char_img = font_img[y1:y2, x1:x2]
    return char_img


def plonk_image(nimbus, background, smaller, coord, custom_size=None):
    """Overlay a smaller image on a background image

    This function also handles cases where the smaller image will overhang
    the screen margins.

    Args:
        background (PIL image): The background image
        smaller (PIL image): The smaller image to overlay on the background
        coord (tuple): The (x, y) position of the bottom left corner of the smaller image
        custom_size (tuple): Override actual screen size

    Returns:
        (PIL image): The modified background image

    """

    # Handle screen width or max width:
    if custom_size is None:
        screen_size = nimbus.screen_size
    else:
        screen_size = custom_size

    # Convert coordinates to PIL offsets
    x_offset = coord[0]
    y_offset = screen_size[1] - coord[1]
    bkg_img = background.copy()
    # Will part of the image be off screen?  Crop it and update offsets if so.
    x = coord[0]
    y = coord[1]
    if x < 0:
        # Image overhangs left-hand side so crop left-hand side of image and set x_offset to 0
        smaller = smaller[:, (-1 * x):, :]
        x_offset = 0
    if y < 0:
        # Image overhangs bottom so crop bottom of image
        # (because PIL has y upside down)
        smaller = smaller[:-((-1*y)), :, :]
        y_offset = screen_size[1]
    if x + smaller.shape[1] > screen_size[0]:
        # right-hand side of image overhangs left-hand side of screen so
        # crop right-hand side of image
        overhanging_length = (x + smaller.shape[1]) - screen_size[0]
        smaller = smaller[:, :-overhanging_length, :]
    if y + smaller.shape[0] > screen_size[1]:
        # top of image overhangs top of screen so crop top of image
        overhanging_length = (y + smaller.shape[0]) - screen_size[1]
        smaller = smaller[overhanging_length:, :, :]
    if x > screen_size[0] or y > screen_size[1]:
        # Image is beyond right-hand side or top so is invisble --> plot nothing
        return bkg_img
    # Overlay the smaller image and return
    bkg_img[y_offset-smaller.shape[0]:y_offset, x_offset:x_offset+smaller.shape[1]] = smaller
    return bkg_img


def colour_char(nimbus, colour, char_img):
    """Colourise a char image (black background, white char)

    Args:
        colour (int): Nimbus colour code

    Returns:
        (PIL image): Colourised char image

    """

    char_img_colourised = char_img.copy()
    white = np.array([255, 255, 255])
    mask = cv2.inRange(char_img, white, white)
    colour_bgr = colour_to_bgr(nimbus, colour)
    b = colour_bgr[0]
    g = colour_bgr[1]
    r = colour_bgr[2]
    char_img_colourised[mask>0] = (b, g, r)
    return char_img_colourised

def plonk_transparent_image(nimbus, background, smaller, coords, is_black=False):
    """Overlay a smaller image on a background with black as transparent

    Args:
        background (PIL image): The background image
        smaller (PIL image): The smaller image to overlay on the background (black=transparent)
        coord (tuple): The (x, y) position of the bottom left corner of the smaller image
        is_black (bool): Temporary workaround
    
    Returns:
        (PIL image): The modified background
    
    """

    # Make an overlay    
    overlay = np.zeros((nimbus.screen_size[1]+1, nimbus.screen_size[0]+1, 3), dtype=np.uint8)
    overlay = plonk_image(nimbus, overlay, smaller, coords)
    
    # Create mask
    gray = cv2.cvtColor(overlay, cv2.COLOR_BGR2GRAY).astype(np.int32)
    gray = np.multiply(gray, 255).clip(max=255).astype(np.uint8)
    mask = cv2.cvtColor(gray, cv2.COLOR_GRAY2BGR)

    # Cast images to signed int32 so we can go negative
    background = background.astype(np.int32)

    # Subtract the mask  from background
    subtracted = np.subtract(background, mask)

    # Set negatives to zero and recast back to uint8
    subtracted = np.clip(subtracted, 0, 255).astype(np.uint8)

    if is_black:
        return subtracted
    else:
        return np.add(subtracted, overlay)